"""
MIT License

Copyright (c) 2021-Present CortexPE

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

from . import InteractMessage, InteractContext, command_map

import typing
from datetime import datetime

import discord
from discord.ext import commands
from discord.http import Route

from discord.types.user import User as UserPayload


class SlashCommands(commands.Cog):
    # TODO: Map commands / subcommand parsing => callback with conversion support to prevent hackily faking a message
    SLASH_INVOKE_PREFIX = "\N{ZERO WIDTH SPACE}"

    def __init__(self, bot: commands.Bot):
        self.registered = []
        self.bot = bot
        self.orig_prefix_callback = bot.get_prefix
        bot.get_prefix = self.get_prefix_wrapper

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

        bot.loop.create_task(self.wait_to_register())

    async def get_prefix_wrapper(self, msg: discord.Message) -> typing.Union[list, str]:
        if not isinstance(msg, InteractMessage):
            return await self.orig_prefix_callback(msg)
        return f"{self.SLASH_INVOKE_PREFIX}/"

    async def wait_to_register(self):
        await self.bot.wait_until_ready()

        if isinstance(self.bot, commands.AutoShardedBot):
            if 0 not in self.bot.shard_ids:  # do not register commands as we are probably running on a separate instance
                return
        elif self.bot.shard_count > 1 and self.bot.shard_id != 0:
            return

        to_send = await self.generate_command_map()
        self.logger.debug("Sending command list...")
        # print(json.dumps(to_send, indent=4))
        await self.register_commands("/", to_send)
        self.logger.info(f"Registered {len(to_send)} top-level commands!")
        # todo: guild-specific commands, refactor this registration code to allow flexible registration
        # await self.register_commands("/guilds/123456789123456789", to_send)

    async def generate_command_map(self) -> typing.List[typing.Dict]:
        cmd_map = []
        for cmd in self.bot.commands:  # type: commands.Command
            cmd_data = await command_map.index_command(self.bot, cmd)
            if cmd_data is None:
                continue
            cmd_map.append(cmd_data)
        return cmd_map

    async def register_commands(self, endpoint: str, cmd_list: list):
        await self.bot.http.request(Route(
            method="PUT", path=f"/applications/{self.bot.application_id}{endpoint}/commands"
        ), json=cmd_list)
        self.registered.append(endpoint)

    def cog_unload(self):
        async def clear_commands():
            self.logger.info(f"Unregistering commands...")
            for endpoint in self.registered:
                await self.bot.http.request(Route(
                    method="PUT", path=f"/applications/{self.bot.application_id}{endpoint}/commands"
                ), json=[])
            self.registered = []

        self.bot.loop.create_task(clear_commands())
        self.bot.get_prefix = self.orig_prefix_callback

    @commands.Cog.listener()
    async def on_interaction(self, interaction: discord.Interaction):
        data = interaction.data
        if "type" not in data:
            return
        if data["type"] != 1:  # chat input cmd type
            return
        cmd = self.bot.get_command(data["name"])
        if cmd is None:
            return

        try:
            ch = self.bot.get_channel(interaction.channel.id) or await self.bot.fetch_channel(interaction.channel.id)
        except discord.HTTPException:
            return await interaction.response.send_message(
                "\N{WARNING SIGN} I cannot access the current channel you are on, please check permissions", ephemeral=True
            )
        if not isinstance(ch, (discord.TextChannel, discord.Thread, discord.DMChannel, discord.GroupChannel, discord.PartialMessageable)):
            return

        if not ch.permissions_for(ch.guild.me if ch.guild is not None else self.bot.user).read_messages:
            # somehow replying is possible even without perms based on my tests, cuz interactions use webhooks instead of actual channel messages,
            # but this might be unwanted behavior.
            return await interaction.response.send_message(
                "\N{WARNING SIGN} I cannot access the current channel you are on, please check permissions", ephemeral=True
            )

        args = []
        for opt in data.get("options", []):
            if opt.get("type", None) == 1:  # chat input cmd type
                args.append(opt["name"])
                for sub_opt in opt.get("options", []):  # type: dict
                    args.append(sub_opt["value"])
                continue
            args.append(opt["value"])

        user = interaction.user

        await interaction.response.defer()

        # fake message, invoke a command as the user
        ctx = await self.bot.get_context(cls=InteractContext, message=InteractMessage(channel=ch, data={
            "content": f"{self.SLASH_INVOKE_PREFIX}/{data['name']} {' '.join(args)}".strip(),
            "author": UserPayload(id=user.id, username=user.name, discriminator=user.discriminator, avatar=None),
            "id": 0,
            "attachments": [],
            "embeds": [],
            "edited_timestamp": datetime.isoformat(datetime.utcnow()),
            "type": discord.MessageType.default,
            "pinned": False,
            "mention_everyone": False,
            "tts": False,
        }, state=self.bot._connection, parent_interaction=interaction))
        await self.bot.invoke(ctx)



"""
MIT License

Copyright (c) 2021-Present CortexPE

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
import typing

import discord
from discord.ext import commands


def param_annotation_to_discord_int(annotation: type):
    if annotation == str or annotation == commands.clean_content:
        return 3
    elif annotation == int:
        return 4
    elif annotation == bool:
        return 5
    elif annotation == discord.user.User:  # user
        return 6
    elif annotation == discord.abc.GuildChannel:  # channel
        return 7
    elif annotation == discord.Role:  # role
        return 8
    # elif False:  # mentionable
    #     return 9
    elif annotation == float:  # number
        return 10

    return 3  # fallback to string


def index_callback_parameters(callback: callable) -> typing.List[dict]:
    options = []
    params = inspect.signature(callback).parameters  # type: typing.Mapping[str, inspect.Parameter]
    for param_name, param in params.items():
        if param_name in ["self", "ctx", "cog"]:
            continue
        # print(param.name, param.annotation, param.empty, param.kind)
        options.append({
            "name": param.name,
            "description": f"{param.name} argument",  # todo: detect parameter descriptions from pydoc
            "type": param_annotation_to_discord_int(param.annotation),
            "required": param.default == param.empty
        })
    return options


async def get_owner_ids(bot: commands.Bot) -> typing.Set[int]:
    ids = set()
    if bot.owner_id:
        ids.add(bot.owner_id)
    elif bot.owner_ids:
        ids = ids.union(set(bot.owner_ids))
    else:
        app = await bot.application_info()  # type: ignore
        if app.team:
            ids = ids.union({m.id for m in app.team.members})
        else:
            ids.add(app.owner.id)
    return ids


async def index_command(bot: commands.Bot, cmd: commands.Command) -> typing.Optional[dict]:
    cmd_data = {
        "name": cmd.name,
        "description": cmd.description or "Un-described command",  # perhaps this needs to be truncated to < 100 chars
        "options": []
    }
    if isinstance(cmd, commands.Group):
        for sub_cmd in cmd.commands:  # type: commands.Command
            sub_cmd_data = await index_command(bot, sub_cmd)
            if sub_cmd_data is None:
                continue
            cmd_data["options"].append(sub_cmd_data)
        if len(cmd_data["options"]) < 1:
            return None
        return cmd_data

    if cmd.hidden:
        return None

    for check in cmd.checks:
        extra_data = check.__dict__.get("slash_extras", None)
        if extra_data is None:
            continue
        if "is_nsfw" in extra_data:
            # todo: discord's slash commands don't support NSFW gating yet
            return None
        if "is_owner" in extra_data:
            return None
            # todo: not sure why this doesn't even work
            # cmd_data["default_permission"] = False
            # cmd_data["permissions"] = [
            #     {
            #         "id": str(owr_id),
            #         "type": 2,  # USER Application Command Permission Type
            #         "permission": True,
            #     } for owr_id in await get_owner_ids(bot)
            # ]

    cmd_data["type"] = 1  # CHAT_INPUT
    if "default_permission" not in cmd_data:
        cmd_data["default_permission"] = True
    if "permissions" not in cmd_data:
        cmd_data["permissions"] = []
    cmd_data["options"] = index_callback_parameters(cmd.callback)

    return cmd_data



from setuptools import setup

setup(
    name="slash_cog",
    version="0.0.1",
    description="A drop-in vanilla discord.py cog to add slash command support with little to no code modifications",
    url="https://github.com/CortexPE/slash_cog",
    author="CortexPE",
    license="MIT",
    packages=["slash_cog"],
    zip_safe=False,
    python_requires=">=3.6",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Typing :: Typed",
    ]
)
